import pydantic


MAX_REPR_LEN = 50


class BaseMessage(pydantic.BaseModel):
    model_config = dict(extra="forbid")

    def serialize(self) -> str:
        """Serialize the data into a JSON string.

        Returns:
            The model data as a JSON string.
        """
        return self.model_dump_json()

    @classmethod
    def deserialize(cls, data: str):
        """Instantiate a model from JSON data.

        Args:
            data: A JSON string.

        Returns:
            An instantiation of the model using the JSON data.
        """
        return cls.model_validate_json(data)

    @staticmethod
    def _shorten(data):
        if isinstance(data, str):
            if len(data) <= MAX_REPR_LEN:
                return repr(data)
            else:
                return f"'{data[:MAX_REPR_LEN]}...'"
        elif isinstance(data, list):
            return f"[{', '.join([BaseMessage._shorten(el) for el in data])}]"
        elif isinstance(data, tuple):
            return f"({', '.join([BaseMessage._shorten(el) for el in data])})"
        elif isinstance(data, dict):
            return f"{{{', '.join([f'{BaseMessage._shorten(key)}: {BaseMessage._shorten(val)}' for key, val in data.items()])}}}"
        else:
            return repr(data)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self)})"

    def __str__(self):
        return ", ".join(
            [f"{prop}={self._shorten(data)}" for prop, data in self.__dict__.items()]
        )



import logging
import os
import socket
import time
from importlib.metadata import entry_points
from typing import Callable, Dict

import stomp
import stomp.exception
from pydantic import ValidationError
from stomp.utils import Frame

from . import messages
from . import exceptions
from .utils import call_with_correct_args, get_version
from threading import Thread


def get_str_time_ms():
    return str(int(time.time_ns() / 1e6))


class Pigeon:
    """A STOMP client with message definitions via Pydantic

    This class is a STOMP message client which will automatically serialize and
    deserialize message data using Pydantic models. Before sending or receiving
    messages, topics must be "registered", or in other words, have a Pydantic
    model associated with each STOMP topic that will be used. This can be done
    in two ways. One is to use the register_topic(), or register_topics()
    methods. The other is to have message definitions in a Python package
    with an entry point defined in the pigeon.msgs group. This entry point
    should provide a tuple containing a mapping of topics to Pydantic models.
    Topics defined in this manner will be automatically discovered and loaded at
    runtime, unless this mechanism is manually disabled.
    """

    def __init__(
        self,
        service: str,
        host: str = "127.0.0.1",
        port: int = 61616,
        logger: logging.Logger = None,
        load_topics: bool = True,
        spawn_threads: bool = False,
        connection_timeout: float = 10,
        send_timeout: float = 0,
    ):
        """
        Args:
            service: The name of the service. This will be included in the
                message headers.
            host: The location of the STOMP message broker.
            port: The port to use when connecting to the STOMP message broker.
            logger: A Python logger to use. If not provided, a logger will be
                crated.
            load_topics: If true, load topics from Python entry points.
            spawn_threads: If true, create a new thread for every message received.
            connection_timeout: The number of seconds to attempt to connect to the broker.
                Set to None to attempt to connect indefinitely.
            send_timeout: The number of seconds to attempt to send a message, or None to
                try indefinitely.

        """
        self._service = service
        self._spawn_threads = spawn_threads
        self._connection = stomp.Connection12([(host, port)], heartbeats=(10000, 10000))
        self._connected = False
        self._connection_timeout = connection_timeout
        self._send_timeout = send_timeout
        self._username = None
        self._password = None
        self._topics = {}
        self._topic_versions = {}
        if load_topics:
            self._load_topics()
        self._callbacks: Dict[str, Callable] = {}
        self._connection.set_listener(
            "listener", PigeonListener(self._handle_message, self._handle_reconnect)
        )
        self._logger = logger if logger is not None else logging.getLogger(__name__)

        self._pid = os.getpid()
        self._hostname = socket.gethostname().split(".")[0]
        self._name = f"{self._service}_{self._pid}_{self._hostname}"

    def _load_topics(self):
        for entrypoint in entry_points(group="pigeon.msgs"):
            self.register_topics(entrypoint.load())

    def register_topic(self, topic: str, msg_class: Callable):
        """Register message definition for a given topic.

        Args:
            topic: The topic that this message definition applies to.
            msg_class: The Pydantic model definition of the message.
        """
        self._topics[topic] = msg_class
        self._topic_versions[topic] = get_version(msg_class)

    def register_topics(self, topics: Dict[str, Callable]):
        """Register a number of message definitions for multiple topics.

        Args:
            topics: A mapping of topics to Pydantic model message definitions.
        """
        for topic in topics.items():
            self.register_topic(*topic)

    def connect(self, username: str = None, password: str = None):
        """
        Connects to the STOMP server using the provided username and password.

        Args:
            username (str, optional): The username to authenticate with. Defaults to None.
            password (str, optional): The password to authenticate with. Defaults to None.

        Raises:
            stomp.exception.ConnectFailedException: If the connection to the server fails.
        """
        self._username = username if username is not None else self._username
        self._password = password if password is not None else self._password

        start = time.time()
        while True:
            try:
                self._connection.connect(
                    username=self._username, passcode=self._password, wait=True
                )
                break
            except stomp.exception.ConnectFailedException as e:
                if (
                    self._connection_timeout is None
                    or time.time() - start < self._connection_timeout
                ):
                    self._logger.error(
                        f"Connection failed: {e}. Attempting to reconnect."
                    )
                    time.sleep(1)
                else:
                    raise stomp.exception.ConnectFailedException(
                        f"Could not connect to server: {e}"
                    ) from e

        self._logger.info("Connected to STOMP server.")

        self._connected = True

    def send(self, topic: str, timeout=..., **data):
        """
        Sends data to the specified topic.

        Args:
            topic (str): The topic to send the data to.
            timeout (float, None): The length of time to attempt to send the
                message, or None to try indefinitely. If set to ..., the default,
                the send_timeout value from the client constructor will be used.
            **data: Keyword arguments representing the data to be sent.

        Raises:
            exceptions.NoSuchTopicException: If the specified topic is not defined.
            TimeoutError: If the message was not able to be sent within the required
                timeframe.
        """
        self._send(topic, data, timeout=timeout)

    def _send(self, topic, data, timeout=..., headers={}):
        """
        Sends data to the specified topic.

        Args:
            topic (str): The topic to send the data to.
            data: The data to send.
            timeout (float, None): The length of time to attempt to send the
                message, or None to try indefinitely. If set to ..., the default,
                the send_timeout value from the client constructor will be used.
            headers: Extra headers to send along with the data.

        Raises:
            exceptions.NoSuchTopicException: If the specified topic is not defined.
            TimeoutError: If the message was not able to be sent within the required
                timeframe.
        """
        self._ensure_topic_exists(topic)
        message = self._topics[topic](**data)
        _headers = dict(
            source=self._name,
            service=self._service,
            hostname=self._hostname,
            pid=self._pid,
            sent_at=get_str_time_ms(),
            version=self._topic_versions[topic],
        )
        _headers.update(headers)
        _timeout = self._send_timeout if timeout is ... else timeout
        assert isinstance(_timeout, (int, float, type(None)))
        start = time.time()
        while True:
            error = None
            if self._connected:
                try:
                    self._connection.send(
                        destination=topic, body=message.serialize(), headers=_headers
                    )
                    self._logger.debug(f"Sent data to {topic}: {message}")
                    return
                except (OSError, stomp.exception.ConnectFailedException) as e:
                    error = e
            if _timeout is not None and time.time() - start >= _timeout:
                exception = TimeoutError(
                    f"Could not send message on topic {topic} within {_timeout} seconds."
                )
                if error is not None:
                    raise exception from error
                else:
                    raise exception
            time.sleep(0.2)

    def _ensure_topic_exists(self, topic: str):
        if topic not in self._topics or topic not in self._topic_versions:
            raise exceptions.NoSuchTopicException(f"Topic {topic} not defined.")

    def _handle_reconnect(self):
        self._connected = False
        self._logger.warning("Disconnected from broker. Attempting to reconnect...")
        old_subscriptions = dict(self._callbacks)
        self._callbacks = {}
        time.sleep(1)
        self.connect()
        for topic, callback in old_subscriptions.items():
            self.subscribe(topic, callback)

    def _handle_message(self, message_frame: Frame):
        topic = message_frame.headers["subscription"]
        if topic not in self._topics:
            self._logger.warning(
                f"Received a message on an unregistered topic: {topic}"
            )
            return
        try:
            message_data = self._topics[topic].deserialize(message_frame.body)
        except ValidationError as e:
            version = message_frame.headers.get("version", "[undefined]")
            self._logger.warning(
                f"Failed to deserialize message on topic '{topic}' due to error:\n{e}Installed message version is {self._topic_versions[topic]} and received message version is {version}"
            )
            return
        callback = self._callbacks.get(topic)
        if callback is None:
            self._logger.warning(
                f"No callback for message received on topic '{topic}'."
            )
            return
        if not self._spawn_threads:
            self._run_callback(callback, message_data, topic, message_frame.headers)
        else:
            Thread(
                target=self._run_callback,
                args=(callback, message_data, topic, message_frame.headers),
            ).start()

    def _run_callback(self, callback, message_data, topic, headers):
        try:
            call_with_correct_args(callback, message_data, topic, headers)
        except exceptions.SignatureException as e:
            self._logger.warning(
                f"Callback signature for topic '{topic}' not acceptable. Call failed with error:\n{e}"
            )
        except Exception as e:
            self._logger.warning(
                f"Callback for topic '{topic}' failed with error:", exc_info=True
            )

    def subscribe(self, topic: str, callback: Callable):
        """
        Subscribes to a topic and associates a callback function to handle incoming messages.

        Args:
            topic (str): The topic to subscribe to.
            callback (Callable): The callback function to handle incoming
                messages. It may accept up to three arguments. In order, the
                arguments are, the received message, the topic the message was
                received on, and the message headers.

        Raises:
            NoSuchTopicException: If the specified topic is not defined.
        """
        self._ensure_topic_exists(topic)
        if topic not in self._callbacks:
            self._connection.subscribe(destination=topic, id=topic)
        self._callbacks[topic] = callback
        self._logger.info(f"Subscribed to {topic} with {callback}.")

    def subscribe_all(self, callback: Callable):
        """Subscribes to all registered topics.

        Args:
            callback: The function to call when a message is received. It must
                accept two arguments, the topic and the message data.
            include_core (bool): If true, subscribe all will subscribe the client to core messages.
        """

        if len(self._topics) == 0:
            self._logger.warning("No topics registered!")
        for topic in self._topics:
            self.subscribe(topic, callback)

    def unsubscribe(self, topic: str):
        """Unsubscribes from a given topic.

        Args:
            topic: The topic to unsubscribe from."""
        self._ensure_topic_exists(topic)
        self._connection.unsubscribe(id=topic)
        self._logger.info(f"Unsubscribed from {topic}.")
        del self._callbacks[topic]

    def disconnect(self):
        """Disconnect from the STOMP message broker."""
        if self._connection.is_connected():
            self._connection.disconnect()
            self._logger.info("Disconnected from STOMP server.")


class PigeonListener(stomp.ConnectionListener):
    def __init__(self, message_handler: Callable, disconnect_handler: Callable):
        self.message_handler = message_handler
        self.disconnect_handler = disconnect_handler

    def on_message(self, frame):
        frame.headers["received_at"] = get_str_time_ms()
        self.message_handler(frame)

    def on_disconnected(self):
        self.disconnect_handler()



from pigeon import utils
from pigeon.exceptions import SignatureException
import pytest
from unittest import mock
import os
import logging
from multiprocessing.queues import Queue


def test_not_enough_args():
    def test_func(a, b, c, d):
        return a, b, c, d

    with pytest.raises(SignatureException):
        utils.call_with_correct_args(test_func, 1, 2, 3)


def test_equal_args():
    def test_func(a, b, c, d):
        return a, b, c, d

    assert utils.call_with_correct_args(test_func, 1, 2, 3, 4) == (1, 2, 3, 4)


def test_args():
    def test_func(a, b, c, d):
        return a, b, c, d

    assert utils.call_with_correct_args(test_func, 1, 2, 3, 4, 5) == (1, 2, 3, 4)


def test_not_enough_kwargs():
    def test_func(a=1, b=2, c=3):
        return a, b, c

    assert utils.call_with_correct_args(test_func, a=10, b=11) == (10, 11, 3)


def test_no_args():
    def test_func():
        return True

    assert utils.call_with_correct_args(test_func, 1, 2, 3)


def test_both():
    def test_func(a, b, c, d=1, e=2):
        return a, b, c, d, e

    assert utils.call_with_correct_args(test_func, 1, 2, 3, 4, 5, d=10, e=11, f=12) == (
        1,
        2,
        3,
        10,
        11,
    )


def test_var_args():
    def test_func(a, b, *args):
        return a, b, args

    assert utils.call_with_correct_args(test_func, 1, 2, 3, 4) == (1, 2, (3, 4))


def test_var_kwargs():
    def test_func(a=1, b=2, **kwargs):
        return a, b, kwargs

    assert utils.call_with_correct_args(test_func, 1, 2, 3, a=10, c=11, d=12) == (
        10,
        2,
        {"c": 11, "d": 12},
    )


def test_both_var():
    def test_func(a, b, *args, c=1, d=2, **kwargs):
        return a, b, c, d, args, kwargs

    assert utils.call_with_correct_args(test_func, 1, 2, 3, 4, e=1, c=12, f=13) == (
        1,
        2,
        12,
        2,
        (3, 4),
        {"e": 1, "f": 13},
    )


def patch_env_vars(**vars):
    return mock.patch.dict(os.environ, **vars)


@pytest.fixture
def mock_loki(mocker):
    return mocker.patch("pigeon.utils.LokiQueueHandler")


def test_setup_logging_basic():
    logger = utils.setup_logging("test_logger")
    assert logger.level == logging.INFO
    assert len(logger.handlers) == 1
    assert logger.handlers[0].level == logging.NOTSET
    assert isinstance(logger.handlers[0], logging.StreamHandler)


def test_setup_logging_root():
    logger = utils.setup_logging()

    assert logger is logging.root


@pytest.mark.parametrize(
    "vars,tags,auth",
    [
        (dict(LOKI_URL="http://a.test.url:1234/", LOKI_VERSION="1"), None, None),
        (
            dict(
                LOKI_URL="https://a.test.secure.url:4321/",
                LOKI_USERNAME="user",
                LOKI_VERSION="2",
            ),
            None,
            ("user", None),
        ),
        (
            dict(LOKI_URL="a.url", LOKI_USERNAME="me", LOKI_PASSWORD="passed"),
            None,
            ("me", "passed"),
        ),
        (
            dict(LOKI_URL="a.url", LOKI_TAGS="tag one: one, tag two : three"),
            {"tag one": "one", "tag two": "three"},
            None,
        ),
    ],
)
def test_setup_logging_loki(request, vars, tags, auth, mock_loki):
    with patch_env_vars(**vars):
        logger = utils.setup_logging(f"loki_test_{request.node.name}", logging.WARN)
        mock_loki.assert_called_once()
        assert len(mock_loki.mock_calls[0].args) == 1
        assert isinstance(mock_loki.mock_calls[0].args[0], Queue)
        assert mock_loki.mock_calls[0].kwargs == dict(
            url=vars.get("LOKI_URL"),
            tags=tags,
            auth=auth,
            version=vars.get("LOKI_VERSION", "1"),
        )
        assert logger.level == logging.WARN
        assert logger.handlers[0].level == logging.NOTSET
        assert len(logger.handlers) == 2
        assert isinstance(logger.handlers[0], logging.StreamHandler)
        assert logger.handlers[1] == mock_loki()


@pytest.mark.parametrize(
    "var, levels",
    [
        ("test.module.one=DEBUG", {"test.module.one": 10}),
        (
            "another.module=INFO,that.thing=WARNING",
            {"another.module": 20, "that.thing": 30},
        ),
        (
            " the.last.module = CRITICAL , the.final.package = ERROR ",
            {"the.last.module": 50, "the.final.package": 40},
        ),
    ],
)
def test_setup_logging_levels(var, levels):
    with patch_env_vars(LOG_LEVEL=var):
        utils.setup_logging("")
        for logger, level in levels.items():
            assert logging.getLogger(logger).level == level



import pytest
from unittest.mock import MagicMock, patch
from stomp.exception import ConnectFailedException
from pigeon.client import Pigeon
from pigeon.exceptions import NoSuchTopicException
from pigeon import BaseMessage
import time
from math import floor


class MockMessage(BaseMessage):
    field1: str


@pytest.fixture
def pigeon_client():
    with patch("pigeon.utils.setup_logging") as mock_logging:
        topics = {"topic1": MockMessage, "topic2": MockMessage}
        client = Pigeon(
            "test",
            host="localhost",
            port=61613,
            logger=mock_logging.Logger(),
            load_topics=False,
            connection_timeout=0.5,
        )
        client.register_topics(topics)
        yield client


@pytest.mark.parametrize(
    "username, password, expected_log",
    [
        (None, None, "Connected to STOMP server."),
        ("user", "pass", "Connected to STOMP server."),
    ],
    ids=["no-auth", "with-auth"],
)
def test_connect(pigeon_client, username, password, expected_log):
    # Arrange
    pigeon_client._connection.connect = MagicMock()
    pigeon_client.subscribe = MagicMock()

    assert not pigeon_client._connected

    # Act
    pigeon_client.connect(username=username, password=password)

    # Assert
    pigeon_client._connection.connect.assert_called_with(
        username=username, passcode=password, wait=True
    )
    pigeon_client._logger.info.assert_called_with(expected_log)
    pigeon_client.subscribe.assert_not_called()

    assert pigeon_client._connected


@pytest.mark.parametrize(
    "username, password",
    [
        (None, None),
        ("user", "pass"),
    ],
    ids=["connect-fail-no-auth", "connect-fail-with-auth"],
)
def test_connect_failure(pigeon_client, username, password):
    # Arrange
    pigeon_client._connection.connect = MagicMock(
        side_effect=ConnectFailedException("Connection failed")
    )
    pigeon_client._logger.error = MagicMock()

    # Act & Assert
    with pytest.raises(
        ConnectFailedException, match="Could not connect to server: Connection failed"
    ):
        pigeon_client.connect(username=username, password=password)

    # Assert the logger was called the same number of times as the retry limit
    assert pigeon_client._logger.error.call_count == 1


@pytest.mark.parametrize(
    "topic, data, expected_serialized_data",
    [
        ("topic1", {"field1": "value"}, '{"field1":"value"}'),
    ],
    ids=["send-data"],
)
def test_send(pigeon_client, topic, data, expected_serialized_data):
    expected_headers = {
        "source": pigeon_client._name,
        "service": "test",
        "hostname": pigeon_client._hostname,
        "pid": pigeon_client._pid,
        "version": pigeon_client._topic_versions[topic],
        "sent_at": "1",
    }
    # Arrange
    with patch("pigeon.client.time.time_ns", lambda: 1e6):
        pigeon_client._topics[topic] = MockMessage
        pigeon_client._connection.send = MagicMock()
        pigeon_client._connected = True

        # Act
        pigeon_client.send(topic, **data)

        # Assert
        pigeon_client._connection.send.assert_called_with(
            destination=topic, body=expected_serialized_data, headers=expected_headers
        )
        pigeon_client._logger.debug.assert_called_with(
            f"Sent data to {topic}: {MockMessage(**data)}"
        )


def test_send_headers(pigeon_client):
    pigeon_client.register_topic("test", MockMessage)
    pigeon_client._connection.send = MagicMock()
    pigeon_client._connected = True

    with patch("pigeon.client.time.time_ns", lambda: 1e6):
        pigeon_client._send("test", {"field1": "datas"}, headers={"test1": "this", "test2": "that"})

    pigeon_client._connection.send.assert_called_with(
        destination="test",
        body='{"field1":"datas"}',
        headers={
            "source": pigeon_client._name,
            "service": "test",
            "hostname": pigeon_client._hostname,
            "pid": pigeon_client._pid,
            "version": pigeon_client._topic_versions["test"],
            "sent_at": "1",
            "test1": "this",
            "test2": "that",
        },
    )



@pytest.mark.parametrize(
    "topic, data",
    [
        ("unknown_topic", {"key": "value"}),
    ],
    ids=["send-data-no-such-topic"],
)
def test_send_no_such_topic(pigeon_client, topic, data):
    pigeon_client._connected = True
    # Act & Assert
    with pytest.raises(NoSuchTopicException, match=f"Topic {topic} not defined."):
        pigeon_client.send(topic, **data)


@pytest.mark.timeout(2)
@pytest.mark.parametrize(
    "class_timeout, method_timeout, timeout",
    [(0, ..., 0), (0.5, ..., 0.5), (0, 0.3, 0.3), (None, 0.7, 0.7)],
)
@pytest.mark.parametrize("reason", ["connected", OSError, ConnectFailedException])
def test_send_timeout(
    mocker, pigeon_client, class_timeout, method_timeout, timeout, reason
):
    client = Pigeon(
        "test",
        host="localhost",
        port=61613,
        logger=mocker.MagicMock(),
        load_topics=False,
        connection_timeout=0.5,
        send_timeout=class_timeout,
    )
    client._connected = reason != "connected"
    client._ensure_topic_exists = mocker.MagicMock()
    client._connection = mocker.MagicMock()
    client.register_topic("some_topic", MockMessage)
    if not isinstance(reason, str) and issubclass(reason, Exception):
        client._connection.send.side_effect = reason
    start = time.time()
    with pytest.raises(TimeoutError):
        client.send("some_topic", timeout=method_timeout, field1="data")
    assert abs(time.time() - start - timeout) < 0.2


@pytest.mark.parametrize(
    "topic, callback_name, expected_log",
    [
        ("topic1", "callback", "Subscribed to topic1 with {}."),
    ],
    ids=["subscribe-new-topic"],
)
def test_subscribe(pigeon_client, topic, callback_name, expected_log):
    # Arrange
    pigeon_client._topics[topic] = MockMessage
    callback = MagicMock(__name__=callback_name)
    pigeon_client._connection.subscribe = MagicMock()

    # Act
    pigeon_client.subscribe(topic, callback)

    # Assert
    assert pigeon_client._callbacks[topic] == callback
    pigeon_client._connection.subscribe.assert_called_with(destination=topic, id=topic)
    pigeon_client._logger.info.assert_called_with(expected_log.format(callback))


@pytest.mark.parametrize(
    "topic",
    [
        ("unknown_topic"),
    ],
    ids=["subscribe-no-such-topic"],
)
def test_subscribe_no_such_topic(pigeon_client, topic):
    # Arrange
    callback = MagicMock()

    # Act & Assert
    with pytest.raises(NoSuchTopicException, match=f"Topic {topic} not defined."):
        pigeon_client.subscribe(topic, callback)


def test_subscribe_all(pigeon_client, mocker):
    pigeon_client._connection = mocker.MagicMock()
    pigeon_client.subscribe = mocker.MagicMock()

    callback = mocker.MagicMock()

    pigeon_client.subscribe_all(callback)

    assert len(pigeon_client.subscribe.mock_calls) == 2
    pigeon_client.subscribe.assert_any_call("topic1", callback)
    pigeon_client.subscribe.assert_any_call("topic2", callback)


def test_subscribe_all_no_topics(pigeon_client, mocker):
    pigeon_client._topics = {}
    pigeon_client._topic_versions = {}
    pigeon_client._connection = mocker.MagicMock()
    pigeon_client.subscribe = mocker.MagicMock()
    pigeon_client._connected = True

    callback = mocker.MagicMock()

    pigeon_client.subscribe_all(callback)
    pigeon_client._logger.warning.assert_called_once()


def test_handle_reconnect(pigeon_client, mocker):
    pigeon_client._connected = True

    pigeon_client.register_topic("topic3", MockMessage)
    pigeon_client.register_topic("topic4", MockMessage)

    topic2_cb = mocker.MagicMock()
    topic3_cb = mocker.MagicMock()

    pigeon_client.connect = mocker.MagicMock()
    pigeon_client._connection = mocker.MagicMock()

    pigeon_client.subscribe("topic2", topic2_cb)
    pigeon_client.subscribe("topic3", topic3_cb)

    pigeon_client.subscribe = mocker.MagicMock()

    pigeon_client._handle_reconnect()

    assert not pigeon_client._connected
    assert pigeon_client._callbacks == {}
    assert len(pigeon_client.subscribe.mock_calls) == 2

    pigeon_client.connect.assert_called_once_with()
    pigeon_client.subscribe.assert_any_call("topic2", topic2_cb)
    pigeon_client.subscribe.assert_any_call("topic3", topic3_cb)
    pigeon_client._logger.warning.assert_called_once_with(
        "Disconnected from broker. Attempting to reconnect..."
    )


@pytest.mark.parametrize(
    "topic, expected_log",
    [
        ("topic1", "Unsubscribed from topic1."),
    ],
    ids=["unsubscribe-existing-topic"],
)
def test_unsubscribe(pigeon_client, topic, expected_log):
    # Arrange
    pigeon_client._callbacks[topic] = ["topic1"]
    pigeon_client._connection.unsubscribe = MagicMock()

    # Act
    pigeon_client.unsubscribe(topic)

    # Assert
    assert topic not in pigeon_client._callbacks
    pigeon_client._connection.unsubscribe.assert_called_with(id=topic)
    pigeon_client._logger.info.assert_called_with(expected_log)


def test_disconnect(pigeon_client):
    # Arrange
    pigeon_client._connection.is_connected = MagicMock(return_value=True)
    pigeon_client._connection.disconnect = MagicMock()

    # Act
    pigeon_client.disconnect()

    # Assert
    pigeon_client._connection.disconnect.assert_called_once()
    pigeon_client._logger.info.assert_called_with("Disconnected from STOMP server.")



from pigeon import BaseMessage
import json
from pydantic import ValidationError
import pytest
from typing import List, Mapping


class Msg(BaseMessage):
    x: int
    y: int


class BigMsg(BaseMessage):
    data: str
    other: int


class BiggerMsg(BaseMessage):
    str: str
    obj: BigMsg
    list: List[BigMsg]
    raw_list: List[str]
    map: Mapping[str, BigMsg]
    raw_map: Mapping[str, str]


def test_serialize():
    data = {"x": 1, "y": 2}
    msg = Msg(**data)

    assert json.loads(msg.serialize()) == data


def test_deserialize():
    data = {"x": 3, "y": 4}
    msg = Msg.deserialize(json.dumps(data))

    assert msg.x == data["x"]
    assert msg.y == data["y"]


def test_forbid_extra():
    with pytest.raises(ValidationError):
        Msg(x=1, y=2, z=3)
    with pytest.raises(ValidationError):
        Msg.deserialize(json.dumps({"x": 1, "y": 2, "z": 3}))


def test_mising_values():
    with pytest.raises(ValidationError):
        Msg(x=1)
    with pytest.raises(ValidationError):
        Msg.deserialize(json.dumps({"y": 3}))


@pytest.fixture
def bigger_msg():
    long_str = 10 * "".join(chr(c) for c in range(ord("a"), ord("z") + 1))
    big_msg = BigMsg(data=long_str, other=101)
    return BiggerMsg(
        str=long_str,
        obj=big_msg,
        list=[big_msg],
        raw_list=[long_str],
        map={"data": big_msg},
        raw_map={"data": long_str},
    )


def test_truncate_repr(bigger_msg):
    assert (
        repr(bigger_msg)
        == "BiggerMsg(str='abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...', obj=BigMsg(data='abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...', other=101), list=[BigMsg(data='abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...', other=101)], raw_list=['abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...'], map={'data': BigMsg(data='abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...', other=101)}, raw_map={'data': 'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...'})"
    )


def test_truncate_str(bigger_msg):
    assert (
        str(bigger_msg)
        == "str='abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...', obj=BigMsg(data='abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...', other=101), list=[BigMsg(data='abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...', other=101)], raw_list=['abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...'], map={'data': BigMsg(data='abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...', other=101)}, raw_map={'data': 'abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwx...'}"
    )



from .client import Pigeon
import argparse
import yaml
import json
from os import environ
import jsonschema2md
from rich.console import Console
from rich.markdown import Markdown
import re
from datetime import datetime
from pydantic import BaseModel
from time import time, sleep


jsonschema2md.Parser.current_locale = "en_US"

console = Console()


class Listener:
    def __init__(self, disp_headers, record=None, rate=False, delay=False, quiet=False):
        self.message_received = False
        self.disp_headers = disp_headers
        self.record = record
        self.rate = rate
        self.delay = delay
        self.quiet = quiet
        self.messages = []
        self.last_timestamp = None

    @classmethod
    def _format(cls, data):
        markdown = []
        if isinstance(data, BaseModel):
            data = data.__dict__
        if isinstance(data, dict):
            for key, val in data.items():
                markdown.append(f"* **{key}**:")
                markdown += ["  " + m for m in cls._format(val)]
        elif isinstance(data, (list, tuple)):
            for val in data:
                child = cls._format(val)
                markdown.append(f"* {child[0]}")
                markdown += ["  " + m for m in child[1:]]
        else:
            markdown.append(str(data))
        return markdown

    @classmethod
    def _to_dict(cls, data):
        if isinstance(data, BaseModel):
            data = data.__dict__
        if isinstance(data, dict):
            return {key: cls._to_dict(val) for key, val in data.items()}
        if isinstance(data, (list, tuple)):
            return [cls._to_dict(val) for val in data]
        if isinstance(data, datetime):
            return data.isoformat()
        return data

    def callback(self, msg, topic, headers):
        if self.record is not None:
            self.messages.append(
                {
                    "msg": self._to_dict(msg),
                    "topic": topic,
                    "headers": headers,
                }
            )
        if not self.quiet:
            markdown = [f"# {topic}"]
            markdown += self._format(msg)
            if self.disp_headers:
                markdown.append(f"### Headers")
                markdown += [f"* **{key}**: {val}" for key, val in headers.items()]
            console.print(Markdown("\n".join(markdown)))
        if self.rate:
            if self.last_timestamp is not None:
                try:
                    rate = 1000 / (
                        int(headers.get("sent_at")) - int(self.last_timestamp)
                    )
                    print(f"Rate: {rate} Hz")
                except ValueError as e:
                    console.print("Could not parse timestamp:", style="red")
                    console.print(e, style="red")
            self.last_timestamp = headers.get("sent_at")
        if self.delay:
            delay = int(headers["received_at"]) - int(headers["sent_at"])
            console.print(f"{delay} ms")
        self.message_received = True

    def write(self):
        with open(self.record, "w") as f:
            json.dump(self.messages, f)


def main():
    parser = argparse.ArgumentParser(prog="Pigeon CLI")
    parser.add_argument(
        "--host",
        type=str,
        help="The message broker to connect to. Defaults to the PIGEON_HOST environment variable if set, otherwise 127.0.0.1.",
    )
    parser.add_argument(
        "--port",
        type=int,
        help="The port to use for the connection. Defaults to the PIGEON_PORT environment variable if set, otherwise 61616.",
    )
    parser.add_argument(
        "--username",
        type=str,
        help="The username to use when connecting to the STOMP server. The environment variable PIGEON_USERNAME is used if set.",
    )
    parser.add_argument(
        "--password",
        type=str,
        help="The password to use when connecting to the STOMP server. The environment variable PIGEON_PASSWORD is used if set.",
    )
    subparsers = parser.add_subparsers(dest="command")

    publish = subparsers.add_parser(
        "publish", aliases=["p", "pub", "send"], help="Publish a message."
    )
    publish.add_argument("topic", type=str, help="The topic to publish a message to.")
    publish.add_argument(
        "data", type=str, help="The YAML/JSON formatted data to publish."
    )

    subscribe = subparsers.add_parser(
        "subscribe", aliases=["s", "sub"], help="Subscribe to topics."
    )
    subscribe.add_argument(
        "topic", type=str, nargs="*", help="The topic to subscribe to."
    )
    subscribe.add_argument(
        "-a", "--all", action="store_true", help="Subscribe to all registered topics."
    )
    subscribe.add_argument(
        "-1", "--one", action="store_true", help="Exit after receiving one message."
    )
    subscribe.add_argument(
        "--headers", action="store_true", help="Display headers of received messages."
    )

    list = subparsers.add_parser("list", aliases=["l"], help="List topics.")

    show = subparsers.add_parser(
        "show", aliases=["doc", "documentation"], help="Display message documentation."
    )
    show.add_argument(
        "topic", type=str, nargs="*", help="The topic to display the documentation for."
    )

    record = subparsers.add_parser("record", aliases=["r"], help="Record messages.")
    record.add_argument(
        "-o", "--output", type=str, help="The file to write the messages to."
    )
    record.add_argument(
        "-a", "--all", action="store_true", help="Subscribe to all registered topics."
    )
    record.add_argument("topic", type=str, nargs="*", help="The topic to record.")

    play = subparsers.add_parser("play", help="Playback recorded messages.")
    play.add_argument(
        "-i",
        "--immediate",
        action="store_true",
        help="Publish all messages without waiting.",
    )
    play.add_argument(
        "-r",
        "--rate",
        type=float,
        default=1,
        help="Multiple the publish rate by this factor.",
    )
    play.add_argument("file", type=str, help="The file to playback.")

    rate = subparsers.add_parser(
        "rate", aliases=["hz"], help="Display the rate at which messages are received."
    )
    rate.add_argument("topic", type=str, help="The topic to use for rate calculations.")

    delay = subparsers.add_parser(
        "delay", aliases=["d"], help="Display the latency of each received message."
    )
    delay.add_argument("topic", type=str, help="The topic to measure the latency of.")

    args = parser.parse_args()

    if args.command is None:
        console.print("No subcommand specified!", style="red")
        exit(1)

    connection = Pigeon(
        "CLI",
        environ.get("PIGEON_HOST", "127.0.0.1") if args.host is None else args.host,
        environ.get("PIGEON_PORT", 61616) if args.port is None else args.port,
    )

    match args.command:
        case "list" | "l":
            markdown = ["# Topics"]
            markdown += [f"* {topic}" for topic in connection._topics]
            console.print(Markdown("\n".join(markdown)))
            exit(0)
        case "show" | "doc" | "documentation":
            parser = jsonschema2md.Parser()
            for topic in args.topic:
                if topic not in connection._topics:
                    print(f"Topic {topic} not found!")
                    continue
                schema = connection._topics[topic].model_json_schema()
                schema["title"] = topic
                markdown = "".join(parser.parse_schema(schema))
                markdown = re.sub(r"<a.*?>(.*?)</a>", r"\1", markdown)
                console.print(Markdown(markdown))
            exit(0)

    connection.connect(
        environ.get("PIGEON_USERNAME") if args.username is None else args.username,
        environ.get("PIGEON_PASSWORD") if args.password is None else args.password,
    )

    match args.command:
        case "publish" | "p" | "pub" | "send":
            connection.send(args.topic, **yaml.safe_load(args.data))
            exit(0)
        case "play":
            with open(args.file) as f:
                messages = json.load(f)
            message_start_time = int(messages[0]["headers"]["sent_at"]) / 1000
            playback_start_time = time()
            for message in messages:
                if not args.immediate:
                    playback_time = args.rate * (time() - playback_start_time)
                    message_time = (
                        int(message["headers"]["sent_at"]) / 1000
                    ) - message_start_time
                    if playback_time < message_time:
                        sleep((message_time - playback_time) / args.rate)
                connection.send(message["topic"], **message["msg"])
            exit(0)
        case "subscribe" | "s" | "sub":
            listener = Listener(args.headers)
            for topic in args.topic:
                connection.subscribe(topic, listener.callback)
            if args.all:
                connection.subscribe_all(listener.callback)
        case "rate" | "hz":
            listener = Listener(False, rate=True, quiet=True)
            connection.subscribe(args.topic, listener.callback)
        case "record" | "r":
            output = args.output
            if output is None:
                output = datetime.now().isoformat()
            if not output.lower().endswith(".json"):
                output += ".json"
            listener = Listener(False, record=output, quiet=True)
            for topic in args.topic:
                connection.subscribe(topic, listener.callback)
            if args.all:
                connection.subscribe_all(listener.callback)
        case "delay" | "d":
            listener = Listener(False, delay=True, quiet=True)
            connection.subscribe(args.topic, listener.callback)

    if len(connection._callbacks) <= 1:
        console.print("No subscriptions, exiting...", style="red")
        exit(0)

    try:
        while not (getattr(args, "one", False) and listener.message_received):
            pass
    except KeyboardInterrupt:
        pass
    finally:
        if listener.record is not None:
            console.print(
                f"Writing {len(listener.messages)} messages to `{listener.record}`",
                style="grey50",
            )
            listener.write()
        console.print("exiting...", style="grey50")

    exit(0)


if __name__ == "__main__":
    main()



import os
import time

from pigeon.client import Pigeon
from pigeon.utils import setup_logging
from pigeon import BaseMessage


class TestMsg(BaseMessage):
    data: str


logger = setup_logging("publisher")

host = os.environ.get("ARTEMIS_HOST", "127.0.0.1")
port = int(os.environ.get("ARTEMIS_PORT", 61616))

connection = Pigeon("Publisher", host=host, port=port, logger=logger, load_topics=False)
connection.register_topic("test", TestMsg)
connection.connect(username="admin", password="password")

while True:
    connection.send("test", data="test_data")
    time.sleep(1)



import os

from pigeon.client import Pigeon
from pigeon.utils import setup_logging
from pigeon import BaseMessage

logger = setup_logging("subscriber")

host = os.environ.get("ARTEMIS_HOST", "127.0.0.1")
port = int(os.environ.get("ARTEMIS_PORT", 61616))


class TestMsg(BaseMessage):
    data: str


def handle_test_message(topic, message):
    logger.info(f"Received {topic} message: {message}")


connection = Pigeon(
    "Subscriber",
    host=host,
    port=port,
    logger=logger,
    load_topics=False,
    connection_timeout=None,
)
connection.register_topic("test", TestMsg)
connection.connect(username="admin", password="password")
connection.subscribe("test", handle_test_message)

while True:
    pass



import logging
from logging import _nameToLevel
import inspect
from copy import copy
import hashlib
from typing import Callable
import os
from logging_loki import LokiQueueHandler
from multiprocessing import Queue
from importlib.metadata import packages_distributions, version

from .exceptions import SignatureException


DISTRIBUTIONS = {
    module: packages[0] for module, packages in packages_distributions().items()
}


def setup_logging(logger_name: str = None, log_level: int = logging.INFO):
    logger = logging.root if logger_name is None else logging.getLogger(logger_name)
    logger.setLevel(log_level)
    stream_handler = logging.StreamHandler()
    stream_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    stream_handler.setFormatter(stream_formatter)
    logger.addHandler(stream_handler)

    setup_loki(logger)
    set_log_levels(logger)

    return logger


def setup_loki(logger):
    if "LOKI_URL" not in os.environ:
        return
    logger.info("Initializing Loki log handler.")
    loki_handler = LokiQueueHandler(
        Queue(-1),
        url=os.environ.get("LOKI_URL"),
        tags=(
            dict(
                [val.strip() for val in tag.split(":")]
                for tag in os.environ.get("LOKI_TAGS").split(",")
            )
            if "LOKI_TAGS" in os.environ
            else None
        ),
        auth=(
            (os.environ.get("LOKI_USERNAME"), os.environ.get("LOKI_PASSWORD"))
            if "LOKI_USERNAME" in os.environ or "LOKI_PASSWORD" in os.environ
            else None
        ),
        version=os.environ.get("LOKI_VERSION", "1"),
    )
    loki_formatter = logging.Formatter("%(message)s")
    loki_handler.setFormatter(loki_formatter)
    logger.addHandler(loki_handler)


def set_log_levels(logger):
    if "LOG_LEVEL" not in os.environ:
        return
    levels = [
        [s.strip() for s in level.split("=")]
        for level in os.environ.get("LOG_LEVEL").split(",")
    ]
    for name, level in levels:
        try:
            logging.getLogger(name).setLevel(_nameToLevel[level])
        except KeyError:
            logger.warning(
                f"Cannot set logger '{name}' to level '{level}' as it is not defined."
            )


def call_with_correct_args(func, *args, **kwargs):
    args = copy(args)
    kwargs = copy(kwargs)
    params = inspect.signature(func).parameters

    if True not in [
        param.kind == inspect._ParameterKind.VAR_POSITIONAL for param in params.values()
    ]:
        num_args = len(
            [
                None
                for param in params.values()
                if param.default == param.empty and param.kind != param.VAR_KEYWORD
            ]
        )
        if num_args > len(args):
            raise SignatureException(
                f"Function '{func}' requires {num_args} positional arguments, but only {len(args)} are available."
            )
        args = args[:num_args]

    if True not in [
        param.kind == inspect._ParameterKind.VAR_KEYWORD for param in params.values()
    ]:
        allowed_keys = [key for key, val in params.items() if val.default != val.empty]
        for key in list(kwargs.keys()):
            if key not in allowed_keys:
                del kwargs[key]

    return func(*args, **kwargs)


def get_version(msg_class):
    return DISTRIBUTIONS.get(msg_class.__module__.split(".")[0], "[unknown]")
